from __future__ import absolute_import

from typing import Optional  # noqa: F401

from .ecdsa import (
    ecdsa_raw_recover,
    ecdsa_raw_sign,
    private_key_to_public_key,
)

from eth_keys.backends.base import BaseECCBackend
from eth_keys.datatypes import (  # noqa: F401
    PrivateKey,
    PublicKey,
    Signature,
)


class NativeECCBackend(BaseECCBackend):
    def ecdsa_sign(self, msg_hash, private_key):
        # type: (bytes, PrivateKey) -> Signature
        signature_vrs = ecdsa_raw_sign(msg_hash, private_key.to_bytes())
        signature = self.Signature(vrs=signature_vrs)
        return signature

    def ecdsa_recover(self,
                      msg_hash,  # type: bytes
                      signature  # type: Signature
                      ):
        # type: (...) -> Optional[PublicKey]
        public_key_bytes = ecdsa_raw_recover(msg_hash, signature.vrs)
        public_key = self.PublicKey(public_key_bytes)
        return public_key

    def private_key_to_public_key(self, private_key):
        # type: (PrivateKey) -> PublicKey
        public_key_bytes = private_key_to_public_key(private_key.to_bytes())
        public_key = self.PublicKey(public_key_bytes)
        return public_key



from __future__ import absolute_import

from typing import Optional  # noqa: F401

from eth_keys.datatypes import (  # noqa: F401
    PrivateKey,
    PublicKey,
    Signature,
)
from eth_keys.exceptions import (
    BadSignature,
)

from .base import BaseECCBackend


def is_coincurve_available():
    # type: () -> bool
    try:
        import coincurve  # noqa: F401
    except ImportError:
        return False
    else:
        return True


class CoinCurveECCBackend(BaseECCBackend):
    def __init__(self):
        # type: () -> None
        try:
            import coincurve
        except ImportError:
            raise ImportError("The CoinCurveECCBackend requires the coincurve \
                               library which is not available for import.")
        self.keys = coincurve.keys
        self.ecdsa = coincurve.ecdsa
        super(CoinCurveECCBackend, self).__init__()

    def ecdsa_sign(self,
                   msg_hash,  # type: bytes
                   private_key  # type: PrivateKey
                   ):
        # type: (...) -> Signature
        private_key_bytes = private_key.to_bytes()
        signature_bytes = self.keys.PrivateKey(private_key_bytes).sign_recoverable(
            msg_hash,
            hasher=None,
        )
        signature = self.Signature(signature_bytes)
        return signature

    def ecdsa_recover(self,
                      msg_hash,  # type: bytes
                      signature  # type: Signature
                      ):
        # type: (...) -> Optional[PublicKey]
        signature_bytes = signature.to_bytes()
        try:
            public_key_bytes = self.keys.PublicKey.from_signature_and_message(
                signature_bytes,
                msg_hash,
                hasher=None,
            ).format(compressed=False)[1:]
        except (ValueError, Exception) as err:
            # `coincurve` can raise `ValueError` or `Exception` dependending on
            # how the signature is invalid.
            raise BadSignature(str(err))
        public_key = self.PublicKey(public_key_bytes)
        return public_key

    def private_key_to_public_key(self, private_key):
        # type: (PrivateKey) -> PublicKey
        public_key_bytes = self.keys.PrivateKey(private_key.to_bytes()).public_key.format(
            compressed=False,
        )[1:]
        return self.PublicKey(public_key_bytes)



from typing import Any  # noqa: F401

from eth_keys import datatypes


class BaseECCBackend(object):
    def __init__(self):
        # type: () -> None
        # Mypy cannot detect the type of dynamically computed classes
        # (https://github.com/python/mypy/issues/2477), so we must annotate those with Any
        self.PublicKey = type(
            '{0}PublicKey'.format(type(self).__name__),
            (datatypes.PublicKey,),
            {'_backend': self},
        )  # type: Any
        self.PrivateKey = type(
            '{0}PrivateKey'.format(type(self).__name__),
            (datatypes.PrivateKey,),
            {'_backend': self},
        )  # type: Any
        self.Signature = type(
            '{0}Signature'.format(type(self).__name__),
            (datatypes.Signature,),
            {'_backend': self},
        )  # type: Any

    def ecdsa_sign(self,
                   msg_hash,    # type: bytes
                   private_key  # type: datatypes.PrivateKey
                   ):
        # type: (...) -> datatypes.Signature
        raise NotImplementedError()

    def ecdsa_verify(self,
                     msg_hash,   # type: bytes
                     signature,  # type: datatypes.Signature
                     public_key  # type: datatypes.PublicKey
                     ):
        # type: (...) -> bool
        return self.ecdsa_recover(msg_hash, signature) == public_key

    def ecdsa_recover(self,
                      msg_hash,  # type: bytes
                      signature  # type: datatypes.Signature
                      ):
        # type: (...) -> datatypes.PublicKey
        raise NotImplementedError()

    def private_key_to_public_key(self,
                                  private_key  # type: datatypes.PrivateKey
                                  ):
        # type: (...) -> datatypes.PublicKey
        raise NotImplementedError()



from __future__ import absolute_import

import codecs
import collections
import sys
from typing import (Any, Optional, Text, Tuple)  # noqa: F401

from eth_utils import (
    big_endian_to_int,
    int_to_big_endian,
    is_bytes,
    keccak,
    to_checksum_address,
    to_normalized_address,
)

from eth_keys.utils.address import (
    public_key_bytes_to_address,
)
from eth_keys.utils.numeric import (
    int_to_byte,
)
from eth_keys.utils.padding import (
    pad32,
)

from eth_keys.exceptions import (
    BadSignature,
    ValidationError,
)
from eth_keys.validation import (
    validate_gte,
    validate_integer,
    validate_lt_secpk1n,
    validate_lte,
    validate_private_key_bytes,
    validate_public_key_bytes,
    validate_signature_bytes,
)


# Workaround for import cycles caused by type annotations:
# http://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
MYPY = False
if MYPY:
    from eth_keys.backends.base import BaseECCBackend  # noqa: F401


# Must compare against version_info[0] and not version_info.major to please mypy.
if sys.version_info[0] == 2:
    ByteString = type(
        b'BaseString',
        (collections.Sequence, basestring),  # noqa: F821
        {},
    )  # type: Any
else:
    ByteString = collections.abc.ByteString


class BackendProxied(object):
    _backend = None

    @property
    def backend(self):
        # type: () -> BaseECCBackend
        from eth_keys.backends import get_backend

        if self._backend is None:
            return get_backend()
        else:
            return self._backend

    @classmethod
    def get_backend(cls):
        # type: () -> BaseECCBackend
        from eth_keys.backends import get_backend

        if cls._backend is None:
            return get_backend()
        else:
            return cls._backend


class BaseKey(ByteString, collections.Hashable):
    _raw_key = None  # type: bytes

    def to_hex(self):
        # type: () -> Text
        # Need the 'type: ignore' comment below because of
        # https://github.com/python/typeshed/issues/300
        return '0x' + codecs.decode(codecs.encode(self._raw_key, 'hex'), 'ascii')  # type: ignore

    def to_bytes(self):
        # type: () -> bytes
        return self._raw_key

    def __hash__(self):
        return big_endian_to_int(keccak(self.to_bytes()))

    def __str__(self):
        return self.to_hex()

    def __unicode__(self):
        return self.__str__()

    def __int__(self):
        return big_endian_to_int(self._raw_key)

    def __len__(self):
        return 64

    def __getitem__(self, index):
        return self._raw_key[index]

    def __eq__(self,
               other  # type: Any
               ):
        # type: (...) -> bool
        if hasattr(other, 'to_bytes'):
            return self.to_bytes() == other.to_bytes()
        elif is_bytes(other):
            return self.to_bytes() == other
        else:
            return False

    def __repr__(self):
        return "'{0}'".format(self.to_hex())

    def __index__(self):
        return self.__int__()

    def __hex__(self):
        # type: () -> Text
        if sys.version_info[0] == 2:
            return codecs.encode(self.to_hex(), 'ascii')
        else:
            return self.to_hex()


class PublicKey(BaseKey, BackendProxied):
    def __init__(self, public_key_bytes):
        # type: (bytes) -> None
        validate_public_key_bytes(public_key_bytes)

        self._raw_key = public_key_bytes

    @classmethod
    def from_private(cls, private_key):
        # type: (PrivateKey) -> PublicKey
        return cls.get_backend().private_key_to_public_key(private_key)

    @classmethod
    def recover_from_msg(cls, message, signature):
        # type: (bytes, Signature) -> PublicKey
        message_hash = keccak(message)
        return cls.recover_from_msg_hash(message_hash, signature)

    @classmethod
    def recover_from_msg_hash(cls, message_hash, signature):
        # type: (bytes, Signature) -> PublicKey
        return cls.get_backend().ecdsa_recover(message_hash, signature)

    def verify_msg(self, message, signature):
        # type: (bytes, Signature) -> bool
        message_hash = keccak(message)
        return self.verify_msg_hash(message_hash, signature)

    def verify_msg_hash(self, message_hash, signature):
        # type: (bytes, Signature) -> bool
        return self.backend.ecdsa_verify(message_hash, signature, self)

    #
    # Ethereum address conversions
    #
    def to_checksum_address(self):
        # type: () -> bytes
        return to_checksum_address(public_key_bytes_to_address(self.to_bytes()))

    def to_address(self):
        # type: () -> bytes
        return to_normalized_address(public_key_bytes_to_address(self.to_bytes()))

    def to_canonical_address(self):
        # type: () -> bytes
        return public_key_bytes_to_address(self.to_bytes())


class PrivateKey(BaseKey, BackendProxied):
    public_key = None  # type: PublicKey

    def __init__(self, private_key_bytes):
        # type: (bytes) -> None
        validate_private_key_bytes(private_key_bytes)

        self._raw_key = private_key_bytes

        self.public_key = self.backend.private_key_to_public_key(self)

    def sign_msg(self, message):
        # type: (bytes) -> Signature
        message_hash = keccak(message)
        return self.sign_msg_hash(message_hash)

    def sign_msg_hash(self, message_hash):
        # type: (bytes) -> Signature
        return self.backend.ecdsa_sign(message_hash, self)


class Signature(ByteString, BackendProxied):
    _backend = None
    _v = None  # type: int
    _r = None  # type: int
    _s = None  # type: int

    def __init__(self, signature_bytes=None, vrs=None):
        # type: (Optional[bytes], Optional[Tuple[int, int, int]]) -> None
        if bool(signature_bytes) is bool(vrs):
            raise TypeError("You must provide one of `signature_bytes` or `vrs`")
        elif signature_bytes:
            validate_signature_bytes(signature_bytes)
            try:
                self.r = big_endian_to_int(signature_bytes[0:32])
                self.s = big_endian_to_int(signature_bytes[32:64])
                self.v = ord(signature_bytes[64:65])
            except ValidationError as err:
                raise BadSignature(str(err))
        elif vrs:
            v, r, s, = vrs
            try:
                self.v = v
                self.r = r
                self.s = s
            except ValidationError as err:
                raise BadSignature(str(err))
        else:
            raise TypeError("Invariant: unreachable code path")

    #
    # v
    #
    @property
    def v(self):
        # type: () -> int
        return self._v

    @v.setter
    def v(self, value):
        # type: (int) -> None
        validate_integer(value)
        validate_gte(value, minimum=0)
        validate_lte(value, maximum=1)

        self._v = value

    #
    # r
    #
    @property
    def r(self):
        # type: () -> int
        return self._r

    @r.setter
    def r(self, value):
        # type: (int) -> None
        validate_integer(value)
        validate_gte(value, 0)
        validate_lt_secpk1n(value)

        self._r = value

    #
    # s
    #
    @property
    def s(self):
        # type: () -> int
        return self._s

    @s.setter
    def s(self, value):
        # type: (int) -> None
        validate_integer(value)
        validate_gte(value, 0)
        validate_lt_secpk1n(value)

        self._s = value

    @property
    def vrs(self):
        # type: () -> Tuple[int, int, int]
        return (self.v, self.r, self.s)

    def to_hex(self):
        # type: () -> Text
        # Need the 'type: ignore' comment below because of
        # https://github.com/python/typeshed/issues/300
        return '0x' + codecs.decode(codecs.encode(self.to_bytes(), 'hex'), 'ascii')  # type: ignore

    def to_bytes(self):
        # type: () -> bytes
        return self.__bytes__()

    def __hash__(self):
        return big_endian_to_int(keccak(self.to_bytes()))

    def __bytes__(self):
        # type: () -> bytes
        vb = int_to_byte(self.v)
        rb = pad32(int_to_big_endian(self.r))
        sb = pad32(int_to_big_endian(self.s))
        # FIXME: Enable type checking once we have type annotations in eth_utils
        return b''.join((rb, sb, vb))  # type: ignore

    def __str__(self):
        return self.to_hex()

    def __unicode__(self):
        return self.__str__()

    def __len__(self):
        return 65

    def __eq__(self, other):
        if hasattr(other, 'to_bytes'):
            return self.to_bytes() == other.to_bytes()
        elif is_bytes(other):
            return self.to_bytes() == other
        else:
            return False

    def __getitem__(self, index):
        return self.to_bytes()[index]

    def __repr__(self):
        return "'{0}'".format(self.to_hex())

    def verify_msg(self, message, public_key):
        # type: (bytes, PublicKey) -> bool
        message_hash = keccak(message)
        return self.verify_msg_hash(message_hash, public_key)

    def verify_msg_hash(self, message_hash, public_key):
        # type: (bytes, PublicKey) -> bool
        return self.backend.ecdsa_verify(message_hash, self, public_key)

    def recover_public_key_from_msg(self, message):
        # type: (bytes) -> PublicKey
        message_hash = keccak(message)
        return self.recover_public_key_from_msg_hash(message_hash)

    def recover_public_key_from_msg_hash(self, message_hash):
        # type: (bytes) -> PublicKey
        return self.backend.ecdsa_recover(message_hash, self)

    def __index__(self):
        return self.__int__()

    def __hex__(self):
        # type: () -> Text
        if sys.version_info[0] == 2:
            return codecs.encode(self.to_hex(), 'ascii')
        else:
            return self.to_hex()

    def __int__(self):
        return big_endian_to_int(self.to_bytes())



from cytoolz import (
    curry,
)

from eth_utils import (
    is_bytes,
    is_integer,
)

from eth_keys.constants import (
    SECPK1_N,
)
from eth_keys.exceptions import (
    ValidationError,
)


def validate_integer(value):
    # type: (int) -> None
    if not is_integer(value) or isinstance(value, bool):
        raise ValidationError("Value must be a an integer.  Got: {0}".format(type(value)))


def validate_bytes(value):
    # type: (bytes) -> None
    if not is_bytes(value):
        raise ValidationError("Value must be a byte string.  Got: {0}".format(type(value)))


@curry
def validate_gte(value, minimum):
    # type: (int, int) -> None
    validate_integer(value)
    if value < minimum:
        raise ValidationError(
            "Value {0} is not greater than or equal to {1}".format(
                value, minimum,
            )
        )


@curry
def validate_lte(value, maximum):
    # type: (int, int) -> None
    validate_integer(value)
    if value > maximum:
        raise ValidationError(
            "Value {0} is not less than or equal to {1}".format(
                value, maximum,
            )
        )


validate_lt_secpk1n = validate_lte(maximum=SECPK1_N - 1)


def validate_message_hash(value):
    # type: (bytes) -> None
    validate_bytes(value)
    if len(value) != 32:
        raise ValidationError("Unexpected signature format.  Must be length 65 byte string")


def validate_public_key_bytes(value):
    # type: (bytes) -> None
    validate_bytes(value)
    if len(value) != 64:
        raise ValidationError("Unexpected public key format.  Must be length 64 byte string")


def validate_private_key_bytes(value):
    # type: (bytes) -> None
    validate_bytes(value)
    if len(value) != 32:
        raise ValidationError("Unexpected private key format.  Must be length 32 byte string")


def validate_signature_bytes(value):
    # type: (bytes) -> None
    validate_bytes(value)
    if len(value) != 65:
        raise ValidationError("Unexpected signature format.  Must be length 65 byte string")



from typing import (Any, Optional, Union)  # noqa: F401

from eth_utils import (
    is_string,
)

from eth_keys.backends import (
    BaseECCBackend,
    get_backend,
)
from eth_keys.datatypes import (
    PublicKey,
    PrivateKey,
    Signature,
)
from eth_keys.exceptions import (
    ValidationError,
)
from eth_keys.validation import (
    validate_message_hash,
)


def backend_property_proxy(name):
    @property
    def property_fn(self):
        backend_property = getattr(self.backend, name)
        return backend_property
    return property_fn


class KeyAPI(object):
    backend = None

    def __init__(self, backend=None):
        if backend is None:
            pass
        elif isinstance(backend, BaseECCBackend):
            pass
        elif isinstance(backend, type) and issubclass(backend, BaseECCBackend):
            backend = backend()
        elif is_string(backend):
            backend = get_backend(backend)
        else:
            raise ValueError(
                "Unsupported format for ECC backend.  Must be an instance or "
                "subclass of `eth_keys.backends.BaseECCBackend` or a string of "
                "the dot-separated import path for the desired backend class"
            )

        self.backend = backend

    _backend = None

    @property
    def backend(self):
        if self._backend is None:
            return get_backend()
        else:
            return self._backend

    @backend.setter
    def backend(self, value):
        self._backend = value

    #
    # Proxy method calls to the backends
    #
    # Mypy cannot detect the type of dynamically computed classes
    # (https://github.com/python/mypy/issues/2477), so we must annotate those with Any
    PublicKey = backend_property_proxy('PublicKey')  # type: Any
    PrivateKey = backend_property_proxy('PrivateKey')  # type: Any
    Signature = backend_property_proxy('Signature')  # type: Any

    def ecdsa_sign(self,
                   message_hash,  # type: bytes
                   private_key  # type: Union[PrivateKey, bytes]
                   ):
        # type: (...) -> Optional[Signature]
        validate_message_hash(message_hash)
        if not isinstance(private_key, PrivateKey):
            raise ValidationError(
                "The `private_key` must be an instance of `eth_keys.datatypes.PrivateKey`"
            )
        signature = self.backend.ecdsa_sign(message_hash, private_key)
        if not isinstance(signature, Signature):
            raise ValidationError(
                "Backend returned an invalid signature.  Return value must be "
                "an instance of `eth_keys.datatypes.Signature`"
            )
        return signature

    def ecdsa_verify(self,
                     message_hash,  # type: bytes
                     signature,  # type: Union[Signature, bytes]
                     public_key  # type: Union[PublicKey, bytes]
                     ):
        # type: (...) -> Optional[bool]
        if not isinstance(public_key, PublicKey):
            raise ValidationError(
                "The `public_key` must be an instance of `eth_keys.datatypes.PublicKey`"
            )
        return self.ecdsa_recover(message_hash, signature) == public_key

    def ecdsa_recover(self,
                      message_hash,  # type: bytes
                      signature  # type: Union[Signature, bytes]
                      ):
        # type: (...) -> Optional[PublicKey]
        validate_message_hash(message_hash)
        if not isinstance(signature, Signature):
            raise ValidationError(
                "The `signature` must be an instance of `eth_keys.datatypes.Signature`"
            )
        public_key = self.backend.ecdsa_recover(message_hash, signature)
        if not isinstance(public_key, PublicKey):
            raise ValidationError(
                "Backend returned an invalid public_key.  Return value must be "
                "an instance of `eth_keys.datatypes.PublicKey`"
            )
        return public_key

    def private_key_to_public_key(self, private_key):
        if not isinstance(private_key, PrivateKey):
            raise ValidationError(
                "The `private_key` must be an instance of `eth_keys.datatypes.PrivateKey`"
            )
        public_key = self.backend.private_key_to_public_key(private_key)
        if not isinstance(public_key, PublicKey):
            raise ValidationError(
                "Backend returned an invalid public_key.  Return value must be "
                "an instance of `eth_keys.datatypes.PublicKey`"
            )
        return public_key


# This creates an easy to import backend which will lazily fetch whatever
# backend has been configured at runtime (as opposed to import or instantiation time).
lazy_key_api = KeyAPI(backend=None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)


install_requires=[
    "eth-utils>=0.5.0",
    "cytoolz>=0.8.2",
]


try:
    import typing
except ImportError:
    # python 2 and 3.4 support
    install_requires.append("typing==3.6.2")


setup(
    name='eth-keys',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='0.1.0-beta.2',
    description="""Common API for Ethereum key operations.""",
    long_description_markdown_filename='README.md',
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    url='https://github.com/ethereum/eth-keys',
    include_package_data=True,
    setup_requires=['setuptools-markdown'],
    install_requires=install_requires,
    py_modules=['eth_keys'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)
